@@ -31,9 +31,9 @@ __all__ = (
 
 
 def _central_difference(
-    x,
+    x: jax.Array,
     *,
-    axis,
+    axis: int,
     left_coeffs,
     center_coeffs,
     right_coeffs,
@@ -63,7 +63,7 @@ def _central_difference(
     return jnp.concatenate([left_x, center_x, right_x], axis=axis)
 
 
-def _forward_difference(
+def _central_difference_edge(
     x,
     *,
     axis,
@@ -88,15 +88,66 @@ def _forward_difference(
     return jnp.concatenate([left_x, right_x], axis=axis)
 
 
-@ft.partial(jax.jit, static_argnames=("accuracy", "axis", "derivative"))
+def _forward_difference(
+    x,
+    *,
+    axis,
+    left_coeffs,
+    right_coeffs,
+    left_offsets,
+    right_offsets,
+):
+    size = x.shape[axis]
+    sliced = ft.partial(jax.lax.slice_in_dim, x, axis=axis)
+
+    left_x = sum(
+        coeff * sliced(offset - left_offsets[0], size + (offset - left_offsets[-1]))
+        for offset, coeff in zip(left_offsets, left_coeffs)
+    )
+
+    right_x = sum(
+        coeff * sliced(size + (offset - left_offsets[-1]), size + (offset))
+        for offset, coeff in zip(right_offsets, right_coeffs)
+    )
+
+    return jnp.concatenate([left_x, right_x], axis=axis)
+
+
+def _backward_difference(
+    x,
+    *,
+    axis,
+    left_coeffs,
+    right_coeffs,
+    left_offsets,
+    right_offsets,
+):
+    size = x.shape[axis]
+    sliced = ft.partial(jax.lax.slice_in_dim, x, axis=axis)
+
+    left_x = sum(
+        coeff * sliced(offset, offset - right_offsets[0])
+        for offset, coeff in zip(left_offsets, left_coeffs)
+    )
+
+    right_x = sum(
+        coeff * sliced(offset - right_offsets[0], size + (offset - right_offsets[-1]))
+        for offset, coeff in zip(right_offsets, right_coeffs)
+    )
+
+    return jnp.concatenate([left_x, right_x], axis=axis)
+
+
+@ft.partial(jax.jit, static_argnames=("accuracy", "axis", "derivative", "method"))
 def difference(
-    x: jnp.ndarray,
+    array: jax.Array,
     *,
     axis: int = 0,
     accuracy: int = 1,
-    step_size: float | jnp.ndarray = 1,
+    step_size: float | jax.Array = 1,
     derivative: int = 1,
-) -> jnp.ndarray:
+    method: str = "central",
+) -> jax.Array:
     """Compute the finite difference derivative along a given axis with a given accuracy
     using central difference for interior points and forward/backward difference for boundary points
     Similar to np.gradient, but with the option to specify accuracy, derivative and step size
@@ -108,6 +159,7 @@ def difference(
         accuracy: accuracy order of the gradient. Default is 1
         step_size: step size. Default is 1
         derivative: derivative order of the gradient. Default is 1
+        method: the method to use (forward, central, backward). Default is central
     Returns:
         Finite difference derivative along the given axis
 
@@ -117,8 +169,8 @@ def difference(
         >>> dx, dy = x[1] - x[0], y[1] - y[0]
         >>> X, Y = jnp.meshgrid(x, y, indexing="ij")
         >>> F =  jnp.sin(X) * jnp.cos(Y)
-        >>> dFdX = difference(F, step_size=dx, axis=0, accuracy=3)
-        >>> dFdXdY = difference(dFdX, step_size=dy, axis=1, accuracy=3)
+        >>> dFdX = difference(F, step_size=dx, axis=0, accuracy=3, method="central")
+        >>> dFdXdY = difference(dFdX, step_size=dy, axis=1, accuracy=3, method="central")
 
         # 1d finite difference derivative
         >>> x = jnp.array([1.2, 1.3, 2.2, 3., 4.5, 5.5, 6., 7., 8., 20.])
@@ -142,7 +194,7 @@ def difference(
         # apply backward difference to the last element with accuracy 1
         x_10 = 20.-8. = 12.
     """
-    size = x.shape[axis]
+    size = array.shape[axis]
 
     center_offsets = _generate_central_offsets(derivative, accuracy + 1)
     center_coeffs = generate_finitediff_coeffs(center_offsets, derivative)
@@ -155,9 +207,9 @@ def difference(
 
     # use central difference for interior points and
     # forward/backward difference for boundary points
-    if size >= len(center_offsets):
+    if method == "central" and size >= len(center_offsets):
         return _central_difference(
-            x,
+            array,
             axis=axis,
             left_coeffs=left_coeffs,
             center_coeffs=center_coeffs,
@@ -167,11 +219,32 @@ def difference(
             right_offsets=right_offsets,
         ) / (step_size**derivative)
 
-    # if size < len(center_offsets), use forward/backward
-    # difference for interior points
-    if size >= len(left_offsets):
+    if method == "central" and size >= len(left_offsets):
+        # if size < len(center_offsets), use forward/backward
+        # difference for interior points
+        # if size >= len(left_offsets):
+        return _central_difference_edge(
+            array,
+            axis=axis,
+            left_coeffs=left_coeffs,
+            right_coeffs=right_coeffs,
+            left_offsets=left_offsets,
+            right_offsets=right_offsets,
+        ) / (step_size**derivative)
+
+    if method == "forward" and size >= len(left_offsets):
         return _forward_difference(
-            x,
+            array,
+            axis=axis,
+            left_coeffs=left_coeffs,
+            right_coeffs=right_coeffs,
+            left_offsets=left_offsets,
+            right_offsets=right_offsets,
+        ) / (step_size**derivative)
+
+    if method == "backward" and size >= len(right_offsets):
+        return _backward_difference(
+            array,
             axis=axis,
             left_coeffs=left_coeffs,
             right_coeffs=right_coeffs,
@@ -182,16 +255,18 @@ def difference(
     msg = f"Size of the array along axis {axis} is smaller than the number of points required"
     msg += f" for the accuracy {accuracy} and derivative {derivative} requested"
     msg += f".\nSize must be larger than {len(left_offsets)}, but got {size}"
+    msg += f".\nMethod should be 'central', 'forward', 'backward', but got {method}"
     raise ValueError(msg)
 
 
-@ft.partial(jax.jit, static_argnames=("accuracy"))
+@ft.partial(jax.jit, static_argnames=("accuracy", "method"))
 def gradient(
-    x: jnp.ndarray,
+    array: jax.Array,
     *,
     accuracy: int | tuple[int, ...] = 1,
-    step_size: float | tuple[float, ...] | jnp.ndarray = 1,
-) -> jnp.ndarray:
+    step_size: float | tuple[float, ...] | jax.Array = 1,
+    method: str = "central",
+) -> jax.Array:
     """Compute the ∇F of input array where F is a scalar function of x and
     returns vectors of the same shape as x stacked along the first axis.
 
@@ -199,6 +274,7 @@ def gradient(
         x: input array
         accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
         step_size: step size. Default is 1, can be a tuple for each axis
+        method: the method to use (forward, central, backward). Default is central
 
     Index notation : dF/dxi
 
@@ -213,25 +289,33 @@ def gradient(
         >>> numpy.testing.assert_allclose(dZdx, dZdx_true, atol=1e-4)
         >>> numpy.testing.assert_allclose(dZdy, dZdy_true, atol=1e-4)
     """
-    accuracy = _check_and_return(accuracy, x.ndim, "accuracy")
-    step_size = _check_and_return(step_size, x.ndim, "step_size")
+    accuracy = _check_and_return(accuracy, array.ndim, "accuracy")
+    step_size = _check_and_return(step_size, array.ndim, "step_size")
 
     return jnp.stack(
         [
-            difference(x, accuracy=acc, step_size=step, derivative=1, axis=axis)
+            difference(
+                array,
+                accuracy=acc,
+                step_size=step,
+                derivative=1,
+                axis=axis,
+                method=method,
+            )
             for axis, (acc, step) in enumerate(zip(accuracy, step_size))
         ],
         axis=0,
     )
 
 
-@ft.partial(jax.jit, static_argnames=("accuracy"))
+@ft.partial(jax.jit, static_argnames=("accuracy", "method"))
 def jacobian(
-    x: jnp.ndarray,
+    array: jax.Array,
     *,
     accuracy: int | tuple[int, ...] = 1,
-    step_size: float | tuple[float, ...] | jnp.ndarray = 1,
-) -> jnp.ndarray:
+    step_size: float | tuple[float, ...] | jax.Array = 1,
+    method: str = "central",
+) -> jax.Array:
     """Compute the ∂Fi/∂xj of input array where F is a vector function of x and
     returns vectors of the same shape as x stacked along the first axis.
 
@@ -239,6 +323,7 @@ def jacobian(
         x: input array
         accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
         step_size: step size. Default is 1, can be a tuple for each axis
+        method: the method to use (forward, central, backward). Default is central
 
     Index notation: ∂Fi/∂xj
 
@@ -257,22 +342,27 @@ def jacobian(
         ...    JF_true = jnp.array([[2 * X * Y, X**2], [5*jnp.ones_like(X), jnp.cos(Y)]])
         ...    npt.assert_allclose(JF, JF_true, atol=1e-7)
     """
-    accuracy = _check_and_return(accuracy, x.ndim - 1, "accuracy")
-    step_size = _check_and_return(step_size, x.ndim - 1, "step_size")
+    accuracy = _check_and_return(accuracy, array.ndim - 1, "accuracy")
+    step_size = _check_and_return(step_size, array.ndim - 1, "step_size")
 
     return jnp.stack(
-        [gradient(xi, accuracy=accuracy, step_size=step_size) for xi in x], axis=0
+        [
+            gradient(xi, accuracy=accuracy, step_size=step_size, method=method)
+            for xi in array
+        ],
+        axis=0,
     )
 
 
-@ft.partial(jax.jit, static_argnames=("accuracy", "keepdims"))
+@ft.partial(jax.jit, static_argnames=("accuracy", "method", "keepdims"))
 def divergence(
-    x: jnp.ndarray,
+    array: jax.Array,
     *,
     accuracy: int | tuple[int, ...] = 1,
     step_size: float | tuple[float, ...] = 1,
     keepdims: bool = True,
-) -> jnp.ndarray:
+    method: str = "central",
+) -> jax.Array:
     """Compute the ∇.F of input array where F is a vector field whose components are the first axis of x
     and returns a scalar field
 
@@ -280,6 +370,7 @@ def divergence(
         x: input array where the leading axis is the dimension of the vector field
         accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
         step_size: step size. Default is 1, can be a tuple for each axis
+        method: the method to use (forward, central, backward). Default is central
         keepdims: whether to keep the leading dimension. Default is True.
 
     Index notation: dFi/dxi
@@ -296,11 +387,18 @@ def divergence(
         >>> divZ_true = 2*X + 3*Y**2  # (dF1/dx) + (dF2/dy)
         >>> numpy.testing.assert_allclose(divZ, divZ_true, atol=5e-4)
     """
-    accuracy = _check_and_return(accuracy, x.ndim - 1, "accuracy")
-    step_size = _check_and_return(step_size, x.ndim - 1, "step_size")
+    accuracy = _check_and_return(accuracy, array.ndim - 1, "accuracy")
+    step_size = _check_and_return(step_size, array.ndim - 1, "step_size")
 
     result = sum(
-        difference(x[axis], accuracy=acc, step_size=step, derivative=1, axis=axis)
+        difference(
+            array[axis],
+            accuracy=acc,
+            step_size=step,
+            derivative=1,
+            axis=axis,
+            method=method,
+        )
         for axis, (acc, step) in enumerate(zip(accuracy, step_size))
     )
 
@@ -309,17 +407,20 @@ def divergence(
     return result
 
 
+@ft.partial(jax.jit, static_argnames=("accuracy", "method"))
 def hessian(
-    x: jnp.ndarray,
+    array: jax.Array,
     *,
     accuracy: int | tuple[int, ...] = 2,
-    step_size: float | tuple[float, ...] | jnp.ndarray = 1,
-) -> jnp.ndarray:
+    step_size: float | tuple[float, ...] | jax.Array = 1,
+    method: str = "central",
+) -> jax.Array:
     """Compute hessian of F: R^m -> R
 
     Args:
         x: input array
         accuracy: accuracy order of the gradient. Default is 2, can be a tuple for each axis
+        method: the method to use (forward, central, backward). Default is central
         step_size: step size. Default is 1, can be a tuple for each axis
 
     Index notation: d2F/dxij
@@ -334,47 +435,60 @@ def hessian(
         >>> H_true = jnp.array([[2 * Y, 2 * X], [2 * X, jnp.zeros_like(X)]])
         >>> npt.assert_allclose(H, H_true, atol=1e-7)
     """
-    accuracy = _check_and_return(accuracy, x.ndim, "accuracy")
-    step_size = _check_and_return(step_size, x.ndim, "step_size")
-    axes = tuple(range(x.ndim))
+    accuracy = _check_and_return(accuracy, array.ndim, "accuracy")
+    step_size = _check_and_return(step_size, array.ndim, "step_size")
+    axes = tuple(range(array.ndim))
     F = dict()
 
     # diag
-    for axis in range(x.ndim):
+    for axis in range(array.ndim):
         F[2 * axis] = difference(
-            x,
+            array,
             accuracy=accuracy[axis],
             step_size=step_size[axis],
             axis=axis,
             derivative=2,
+            method=method,
         )
 
     # off-diag
     for ax1, ax2 in combinations(axes, 2):
         F[ax1 + ax2] = difference(
-            difference(x, accuracy=accuracy[ax1], step_size=step_size[ax1], axis=ax1),
+            difference(
+                array,
+                accuracy=accuracy[ax1],
+                step_size=step_size[ax1],
+                method=method,
+                axis=ax1,
+            ),
             accuracy=accuracy[ax2],
             step_size=step_size[ax2],
+            method=method,
             axis=ax2,
         )
 
     return jnp.stack(
-        [jnp.stack([F[ax1 + ax2] for ax2 in range(x.ndim)]) for ax1 in range(x.ndim)]
+        [
+            jnp.stack([F[ax1 + ax2] for ax2 in range(array.ndim)])
+            for ax1 in range(array.ndim)
+        ]
     )
 
 
-@ft.partial(jax.jit, static_argnames=("accuracy"))
+@ft.partial(jax.jit, static_argnames=("accuracy", "method"))
 def laplacian(
-    x: jnp.ndarray,
+    array: jax.Array,
     *,
     accuracy: int | tuple[int, ...] = 1,
-    step_size: float | tuple[float, ...] | jnp.ndarry = 1,
-) -> jnp.ndarray:
+    step_size: float | tuple[float, ...] | jax.Array = 1,
+    method: str = "central",
+) -> jax.Array:
     """Compute the ΔF of input array.
     Args:
         x: input array
         accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
         step_size: step size. Default is 1, can be a tuple for each axis
+        method: the method to use (forward, central, backward). Default is central
 
     Index notation: d2F/dxi2
     Example:
@@ -387,22 +501,25 @@ def laplacian(
         >>> laplacianZ_true = 12*X**2 + 6*Y
         >>> numpy.testing.assert_allclose(laplacianZ, laplacianZ_true, atol=1e-4)
     """
-    accuracy = _check_and_return(accuracy, x.ndim, "accuracy")
-    step_size = _check_and_return(step_size, x.ndim, "step_size")
+    accuracy = _check_and_return(accuracy, array.ndim, "accuracy")
+    step_size = _check_and_return(step_size, array.ndim, "step_size")
 
     return sum(
-        difference(x, accuracy=acc, step_size=step, derivative=2, axis=axis)
+        difference(
+            array, accuracy=acc, step_size=step, derivative=2, method=method, axis=axis
+        )
         for axis, (acc, step) in enumerate(zip(accuracy, step_size))
     )
 
 
 def _curl_2d(
-    x: jnp.ndarray,
+    array: jax.Array,
     *,
     accuracy: int | tuple[int, ...] = 1,
-    step_size: float | tuple[float, ...] | jnp.ndarray = 1,
+    step_size: float | tuple[float, ...] | jax.Array = 1,
+    method: str = "central",
     keepdims: bool = True,
-) -> jnp.ndarray:
+) -> jax.Array:
     """Compute the ∇×F of input array where F is a vector field whose components are the first axis of x
     and returns a vector field
 
@@ -412,6 +529,7 @@ def _curl_2d(
         x: input array where the leading axis is the dimension of the vector field
         accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
         step_size: step size. Default is 1, can be a tuple for each axis
+        method: the method to use (forward, central, backward). Default is central
         keepdims: whether to keep the leading dimension of the vector field
 
     Example:
@@ -423,8 +541,12 @@ def _curl_2d(
         >>> F = jnp.stack([F1,F2], axis=0)
         >>> curl = curl_2d(F, accuracy=4, step_size=dx)
     """
-    dF1dY = difference(x[0], accuracy=accuracy[1], step_size=step_size[1], axis=1)
-    dF2dX = difference(x[1], accuracy=accuracy[0], step_size=step_size[0], axis=0)
+    dF1dY = difference(
+        array[0], accuracy=accuracy[1], step_size=step_size[1], method=method, axis=1
+    )
+    dF2dX = difference(
+        array[1], accuracy=accuracy[0], step_size=step_size[0], method=method, axis=0
+    )
 
     result = dF2dX - dF1dY
 
@@ -435,19 +557,32 @@ def _curl_2d(
 
 
 def _curl_3d(
-    x: jnp.ndarray,
+    array: jax.Array,
     *,
     accuracy: int | tuple[int, ...] = 1,
     step_size: float | tuple[float, ...] | jnp.ndarry = 1,
-) -> jnp.ndarray:
-    dF1dY = difference(x[0], accuracy=accuracy[1], step_size=step_size[1], axis=1)
-    dF1dZ = difference(x[0], accuracy=accuracy[2], step_size=step_size[2], axis=2)
+    method: str = "central",
+) -> jax.Array:
+    dF1dY = difference(
+        array[0], accuracy=accuracy[1], step_size=step_size[1], method=method, axis=1
+    )
+    dF1dZ = difference(
+        array[0], accuracy=accuracy[2], step_size=step_size[2], method=method, axis=2
+    )
 
-    dF2dX = difference(x[1], accuracy=accuracy[0], step_size=step_size[0], axis=0)
-    dF2dZ = difference(x[1], accuracy=accuracy[2], step_size=step_size[2], axis=2)
+    dF2dX = difference(
+        array[1], accuracy=accuracy[0], step_size=step_size[0], method=method, axis=0
+    )
+    dF2dZ = difference(
+        array[1], accuracy=accuracy[2], step_size=step_size[2], method=method, axis=2
+    )
 
-    dF3dX = difference(x[2], accuracy=accuracy[0], step_size=step_size[0], axis=0)
-    dF3dY = difference(x[2], accuracy=accuracy[1], step_size=step_size[1], axis=1)
+    dF3dX = difference(
+        array[2], accuracy=accuracy[0], step_size=step_size[0], method=method, axis=0
+    )
+    dF3dY = difference(
+        array[2], accuracy=accuracy[1], step_size=step_size[1], method=method, axis=1
+    )
 
     return jnp.stack(
         [
@@ -459,14 +594,15 @@ def _curl_3d(
     )
 
 
-@ft.partial(jax.jit, static_argnames=("accuracy", "keepdims"))
+@ft.partial(jax.jit, static_argnames=("accuracy", "method", "keepdims"))
 def curl(
-    x: jnp.ndarray,
+    array: jax.Array,
     *,
     accuracy: int | tuple[int, ...] = 1,
-    step_size: float | tuple[float, ...] | jnp.ndarry = 1,
+    step_size: float | tuple[float, ...] | jax.Array = 1,
+    method: str = "central",
     keepdims: bool = True,
-) -> jnp.ndarray:
+) -> jax.Array:
     """Compute the ∇×F of input array where F is a vector field whose components are the first axis of x
     and returns a vector field
 
@@ -476,6 +612,7 @@ def curl(
         x: input array where the leading axis is the dimension of the vector field
         accuracy: accuracy order of the gradient. Default is 1, can be a tuple for each axis
         step_size: step size. Default is 1, can be a tuple for each axis
+        method: the method to use (forward, central, backward). Default is central
         keepdims: whether to keep the leading dimension of the vector field (only for 2D)
 
     Example:
@@ -490,7 +627,7 @@ def curl(
         >>> F2 = X**4 + Y**3
         >>> F3 = jnp.zeros_like(F1)
         >>> F = jnp.stack([F1,F2,F3], axis=0)
-        >>> curlF = sk.fd.curl(F, step_size=(dx,dy,dz),  accuracy=6)
+        >>> curlF = finitediffx.curl(F, step_size=(dx,dy,dz),  accuracy=6)
         >>> curlF_exact = jnp.stack([F1*0,F1*0, 4*X**3 - 3*Y**2], axis=0)
         >>> npt.assert_allclose(curlF, curlF_exact, atol=1e-7)
 
@@ -504,15 +641,22 @@ def curl(
         >>> curl = curl_2d(F, accuracy=4, step_size=dx)
     """
 
-    accuracy = _check_and_return(accuracy, x.ndim - 1, "accuracy")
-    step_size = _check_and_return(step_size, x.ndim - 1, "step_size")
+    accuracy = _check_and_return(accuracy, array.ndim - 1, "accuracy")
+    step_size = _check_and_return(step_size, array.ndim - 1, "step_size")
 
-    if x.ndim == 4 and x.shape[0] == 3:
-        return _curl_3d(x, accuracy=accuracy, step_size=step_size)
+    if array.ndim == 4 and array.shape[0] == 3:
+        return _curl_3d(array, accuracy=accuracy, method=method, step_size=step_size)
 
-    if x.ndim == 3 and x.shape[0] == 2:
-        return _curl_2d(x, accuracy=accuracy, step_size=step_size, keepdims=keepdims)
+    if array.ndim == 3 and array.shape[0] == 2:
+        return _curl_2d(
+            array,
+            accuracy=accuracy,
+            step_size=step_size,
+            method=method,
+            keepdims=keepdims,
+        )
 
-    msg = f"`curl` is only implemented for 2D and 3D vector fields, got {x.ndim}D"
-    msg += "for 2D vector fields, the leading axis must be 2, for 3D vector fields, the leading axis must be 3"
+    msg = f"`curl` is only implemented for 2D and 3D vector fields, got {array.ndim}D"
+    msg += "for 2D vector fields, the leading axis must have a shape=2, "
+    msg += "for 3D vector fields, the leading axis must have a shape=3"
     raise ValueError(msg)
